exclude(:test_syntax_error_message, "error message format")

exclude(:test_each_child, "https://bugs.ruby-lang.org/issues/20479")
exclude(:test_trace_points, "https://bugs.ruby-lang.org/issues/20479")
